/*************************************************************************
 **
 ** FILE NAME    : qos_mgr_tc.c
 ** PROJECT      : QOS MGR
 ** MODULES      : QOS MGR (TC APIs)
 **
 ** DATE         : 18 APR 2018
 ** AUTHOR       : Purnendu Ghosh
 ** DESCRIPTION  : QOS MGR TC API Implementation
 ** COPYRIGHT : Copyright (c) 2017 - 2018 Intel Corporation
 ** Copyright (c) 2010 - 2016 Lantiq Beteiligungs-GmbH & Co. KG
 **
 *************************************************************************/

#include <net/flow_dissector.h>
#include <net/pkt_cls.h>
#include <net/tc_act/tc_gact.h>
#include <net/tc_act/tc_vlan.h>
#include <net/tc_act/tc_mirred.h>

#include <net/datapath_api_vlan.h>
#include <net/ppa/ppa_api.h>
#include <net/ppa/qos_mgr_tc_hook.h>

PPA_LOCK g_qos_mgr_list_lock;
int32_t g_Rule_status;

#define QOS_MGR_VLAN_NUM 2
#define QOS_MGR_MAX_IFACE 16
#define QOS_MGR_MAX_QUEUE_IFACE 8

struct qos_mgr_match_vlan {
#define QOS_MGR_VLAN_ID		0x01
#define QOS_MGR_VLAN_PRIO	0x02
#define QOS_MGR_VLAN_PROTO	0x04
#define QOS_MGR_VLAN_TPID	0x08
	int32_t			vlan_mask;

	int32_t prio; /* match exact VLAN tag priority: 0-7 */
	int32_t vid; /* match exact VID: 0 - 4093 */
	int32_t tpid; /* match exact TPID: 0x8100 or another configured TPID */
	int32_t dei; /* match exact DEI: 0 or 1 */
	int32_t proto /* match exact ether type (protocol) */
};

struct qos_mgr_act_vlan {
#define QOS_MGR_ACT_DROP   0x01  /*drop packet */
#define QOS_MGR_VLAN_ACT_POP    0x02  /*pop/remove VLAN */
#define QOS_MGR_VLAN_ACT_PUSH   0x04  /*push/insert VLAN */
#define QOS_MGR_VLAN_ACT_MODIFY   0x08  /*modify  VLAN */
	int32_t act;  /* if act == QOS_MGR_ACT_DROP, drop the packet
		   * if act == QOS_MGR_VLAN_ACT_POP, remove VLAN
		   * if act == QOS_MGR_VLAN_ACT_PUSH, insert VLAN
		   */
	int32_t pop_n;  /*the number of VLAN tag to pop
		     *the valid number: 1 or 2
		     */
	int32_t push_n;  /*the number of VLAN tag to push
		      *the valid number: 1 or 2
		      */
#define CP_FROM_IN    -1 /*copy from inner VLAN header*/
#define CP_FROM_OUT   -2 /*copy from inner VLAN header*/
	int32_t tpid[QOS_MGR_VLAN_NUM]; /* the tpid of VLAN to push:
				* support two TPID 0x8100 and
				* another programmable TPID
				* or
				*  copy from recv pkt's inner tag(CP_FROM_IN)
				*  copy from recv pkt's outer tag(CP_FROM_OUT)
				*/
	int32_t vid[QOS_MGR_VLAN_NUM]; /* the VID of VLAN to push:
				*  support range: 0 - 4093
				* or
				*  copy from recv pkt's inner tag(CP_FROM_IN)
				*  copy from recv pkt's outer tag(CP_FROM_OUT)
				*/
	int32_t dei[QOS_MGR_VLAN_NUM]; /* the DEI of VLAN to push:
				*  support range: 0 - 1
				* or
				*  copy from recv pkt's inner tag(CP_FROM_IN)
				*  copy from recv pkt's outer tag(CP_FROM_OUT)
				*  keep existing value
				*/
	int32_t prio[QOS_MGR_VLAN_NUM]; /* the prority of VLAN to push:
				*  support range: 0 - 7
				* or
				*  copy from recv pkt's inner tag(CP_FROM_IN)
				*  copy from recv pkt's outer tag(CP_FROM_OUT)
				*/
};

struct qos_mgr_tc_flow_rule_act {
	int8_t			act_type;
	struct qos_mgr_act_vlan		action; /*action once matched */
};

struct qos_mgr_vlan_entry_match {
	uint32_t			no_of_tag;
	struct qos_mgr_match_vlan	entry[QOS_MGR_VLAN_NUM];/* match pattern.
					* only proto is valid for this case
					*/
};

struct qos_mgr_eth_mac_match {
#define QOS_MGR_ETH_SRC_MAC	0x01
#define QOS_MGR_ETH_DST_MAC	0x02
	uint8_t which_mac;
	uint8_t dst[ETH_ALEN];
	uint8_t src[ETH_ALEN];
};

enum flow_rule_match_to {
	QOS_MGR_MATCH_UNKNOWN,
	QOS_MGR_MATCH_ETH_ADDRS,
	QOS_MGR_MATCH_PROTO,
	QOS_MGR_MATCH_VLAN,
};

struct qos_mgr_tc_flow_rule_match {
	uint32_t		used_keys;
	uint32_t		match_to;
	union {
		struct qos_mgr_vlan_entry_match vlan_match;
		struct qos_mgr_eth_mac_match	eth_match;
	}
};

struct qos_mgr_tc_flow_rule_list_item {
	PPA_ATOMIC                                      count;
	uint32_t					dir;
	uint32_t					status;
	uint32_t					priority;
	uint32_t					proto;
	uint32_t					parentid;
	struct net_device				*indev;
	struct qos_mgr_tc_flow_rule_list_item		*next;
	struct qos_mgr_tc_flow_rule_act			flow_action;
	struct qos_mgr_tc_flow_rule_match		flow_pattern;

};

struct __qos_mgr_tc_queue_info {
	uint32_t p_handle;
	uint32_t priority;
};

struct qos_mgr_tc_queue_info {
	struct __qos_mgr_tc_queue_info q_info[QOS_MGR_MAX_QUEUE_IFACE];
};

struct _tc_qos_mgr_db {
	struct net_device				*dev;
	struct qos_mgr_tc_queue_info			mq_info;
	struct qos_mgr_tc_flow_rule_list_item		*flow_item_ig;
	struct qos_mgr_tc_flow_rule_list_item		*flow_item_eg;
};
/** Right now maximum of 16 CTP/bridge port are considered */
struct _tc_qos_mgr_db tc_qos_mgr_db[QOS_MGR_MAX_IFACE];

static int32_t qos_mgr_set_ext_vlan_to_dp(struct net_device *dev,
			struct qos_mgr_tc_flow_rule_list_item  *fl_list);

void qos_mgr_tc_lock_list(void)
{
	ppa_lock_get(&g_qos_mgr_list_lock);
}

void qos_mgr_tc_unlock_list(void)
{
	ppa_lock_release(&g_qos_mgr_list_lock);
}

struct qos_mgr_tc_flow_rule_list_item *tc_qos_mgr_alloc_list_item(void)
{
	struct qos_mgr_tc_flow_rule_list_item  *obj;

	obj = (struct qos_mgr_tc_flow_rule_list_item  *)ppa_malloc(sizeof(*obj));
	if (obj) {
		ppa_memset(obj, 0, sizeof(*obj));
		ppa_atomic_set(&obj->count, 1);
	}
	return obj;
}

void __qos_mgr_tc_add_list_item(struct qos_mgr_tc_flow_rule_list_item *obj,
			struct qos_mgr_tc_flow_rule_list_item **obj_head)
{
	ppa_atomic_inc(&obj->count);
	obj->next = *obj_head;
	*obj_head = obj;
}

void qos_mgr_tc_free_list_item(struct qos_mgr_tc_flow_rule_list_item *obj)
{
	if (ppa_atomic_dec(&obj->count) == 0)
		ppa_free(obj);
}

void tc_qos_mgr_remove_list_item(uint32_t prio,
		struct qos_mgr_tc_flow_rule_list_item **obj_head)
{
	struct qos_mgr_tc_flow_rule_list_item *p_prev, *p_cur;
	p_prev = NULL;

	qos_mgr_tc_lock_list();
	for (p_cur = *obj_head; p_cur; p_prev = p_cur, p_cur = p_cur->next)
		if ((p_cur->priority == prio)) {
			if (!p_prev)
				*obj_head = p_cur->next;
			else
				p_prev->next = p_cur->next;

			qos_mgr_tc_free_list_item(p_cur);
			break;
		}
	qos_mgr_tc_unlock_list();
}

void qos_mgr_tc_list_item_put(struct qos_mgr_tc_flow_rule_list_item *obj)
{
        if (obj)
                qos_mgr_tc_free_list_item(obj);
}

char *get_pattern_match_type(uint32_t pattern)
{

	switch (pattern) {
	case QOS_MGR_MATCH_UNKNOWN: return "Unknown";
	case QOS_MGR_MATCH_ETH_ADDRS: return "Ethernet MAC";
	case QOS_MGR_MATCH_PROTO: return "Protocol";
	case QOS_MGR_MATCH_VLAN: return "VLAN";
	}
	return "No Match";
}

char *get_action_type(int8_t act_type)
{
	switch (act_type) {
	case TC_ACT_OK: return "OK";
	case TC_ACT_SHOT: return "DROP";
	case TC_ACT_UNSPEC: return "CONTINUE";
	case TCA_ACT_VLAN: return "VLAN";
	case TCA_ACT_MIRRED: return "REDIRECT";
	default: return "UNKNOWN";
	}
}

char *get_dir(uint8_t dir)
{
	switch (dir) {
	case 0: return "EGRESS";
	case 1: return "INGRESS";
	default: return "INVALID";
	}
}

int32_t __qos_mgr_tc_flow_lookup(struct qos_mgr_tc_flow_rule_list_item  *fl_list)
{
	int32_t ret = PPA_ENOTAVAIL;
	struct qos_mgr_tc_flow_rule_list_item *p;

	for (p = fl_list; p; p = p->next) {
		if (p != NULL) {
			/*ppa_atomic_inc(&p->count); */
			pr_debug("=====================================================\n");
			pr_debug("Direction: %s  Priority: %d\n",
					get_dir(p->dir), p->priority);
			pr_debug("PATTERN Used Keys: %d\n",
					p->flow_pattern.used_keys);
			pr_debug("Pattern: %s\n",
					get_pattern_match_type(p->flow_pattern.match_to));
			if (p->flow_pattern.match_to == QOS_MGR_MATCH_VLAN) {
				int32_t i = 0;
				pr_debug("\t Number Of Tag:%d\n", p->flow_pattern.vlan_match.no_of_tag);
				for (i = 0; i < p->flow_pattern.vlan_match.no_of_tag ; i++) {
					pr_debug("\t vlan_mask:%x\n",
						p->flow_pattern.vlan_match.entry[i].vlan_mask);
					pr_debug("\t VID:%d Priority:%d\n",
						p->flow_pattern.vlan_match.entry[i].vid,
						p->flow_pattern.vlan_match.entry[i].prio);
				}
			}
			if (p->flow_pattern.match_to == QOS_MGR_MATCH_ETH_ADDRS) {
				if (p->flow_pattern.eth_match.which_mac == (QOS_MGR_ETH_SRC_MAC | QOS_MGR_ETH_DST_MAC)) {
					pr_debug("\t SRC MAC Addr: %02x:%02x:%02x:%02x:%02x:%02x\n",
							p->flow_pattern.eth_match.src[0],
							p->flow_pattern.eth_match.src[1],
							p->flow_pattern.eth_match.src[2],
							p->flow_pattern.eth_match.src[3],
							p->flow_pattern.eth_match.src[4],
							p->flow_pattern.eth_match.src[5]);
					pr_debug("\t DST MAC Addr: %02x:%02x:%02x:%02x:%02x:%02x\n",
							p->flow_pattern.eth_match.dst[0],
							p->flow_pattern.eth_match.dst[1],
							p->flow_pattern.eth_match.dst[2],
							p->flow_pattern.eth_match.dst[3],
							p->flow_pattern.eth_match.dst[4],
							p->flow_pattern.eth_match.dst[5]);
				} else if (p->flow_pattern.eth_match.which_mac == QOS_MGR_ETH_SRC_MAC) {
					pr_debug("\t SRC MAC Addr: %02x:%02x:%02x:%02x:%02x:%02x\n",
							p->flow_pattern.eth_match.src[0],
							p->flow_pattern.eth_match.src[1],
							p->flow_pattern.eth_match.src[2],
							p->flow_pattern.eth_match.src[3],
							p->flow_pattern.eth_match.src[4],
							p->flow_pattern.eth_match.src[5]);
				} else if (p->flow_pattern.eth_match.which_mac == QOS_MGR_ETH_DST_MAC) {
					pr_debug("\t DST MAC Addr: %02x:%02x:%02x:%02x:%02x:%02x\n",
							p->flow_pattern.eth_match.dst[0],
							p->flow_pattern.eth_match.dst[1],
							p->flow_pattern.eth_match.dst[2],
							p->flow_pattern.eth_match.dst[3],
							p->flow_pattern.eth_match.dst[4],
							p->flow_pattern.eth_match.dst[5]);
				}

			}
			pr_debug("ACTION Type: %s\n", get_action_type(p->flow_action.act_type));
			if (p->flow_action.act_type == TCA_ACT_VLAN) {
				if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_POP) {
					pr_debug("\t POP\n");
					pr_debug("\t No Of POP: %d\n", p->flow_action.action.pop_n);
				} else if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_PUSH) {
					pr_debug("\t PUSH\n");
					pr_debug("\t No Of PUSH: %d\n", p->flow_action.action.push_n);
					pr_debug("\t VID: %02x Prio: %d Proto: %02x\n",
							p->flow_action.action.vid[p->flow_action.action.push_n - 1],
							p->flow_action.action.prio[p->flow_action.action.push_n - 1],
							p->flow_action.action.tpid[p->flow_action.action.push_n - 1]);
				} else if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_MODIFY) {
					pr_debug("\t MODIFY\n");
					pr_debug("\t VID: %d Prio: %d Proto: %02x\n",
							p->flow_action.action.vid[0],
							p->flow_action.action.prio[0],
							p->flow_action.action.tpid[0]);
				}
			}
			if (p->flow_action.act_type == TC_ACT_SHOT) {
				pr_debug("\t DROP\n");
			}
			ret = PPA_SUCCESS;
		}
	}

	return ret;
}

int32_t qos_mgr_tc_flow_lookup(struct qos_mgr_tc_flow_rule_list_item  *fl_list)
{
	int32_t ret;

	qos_mgr_tc_lock_list();
	ret = __qos_mgr_tc_flow_lookup(fl_list);
	qos_mgr_tc_unlock_list();

	return ret;
}

static void tc_flower_store_eth_addr_match(
			struct flow_dissector_key_eth_addrs *key,
			struct flow_dissector_key_eth_addrs *mask,
			struct qos_mgr_tc_flow_rule_list_item *tc_fl_item)
{
	tc_fl_item->flow_pattern.match_to = QOS_MGR_MATCH_ETH_ADDRS;

	if (mask->dst[0] == 0) {
		tc_fl_item->flow_pattern.eth_match.which_mac = QOS_MGR_ETH_SRC_MAC;
		memcpy(tc_fl_item->flow_pattern.eth_match.src, key->src, ETH_ALEN);
	} else if (mask->src[0] == 0) {
		tc_fl_item->flow_pattern.eth_match.which_mac = QOS_MGR_ETH_DST_MAC;
		memcpy(tc_fl_item->flow_pattern.eth_match.dst, key->dst, ETH_ALEN);
	} else {
		tc_fl_item->flow_pattern.eth_match.which_mac = QOS_MGR_ETH_SRC_MAC | QOS_MGR_ETH_DST_MAC;
		memcpy(tc_fl_item->flow_pattern.eth_match.src, key->src, ETH_ALEN);
		memcpy(tc_fl_item->flow_pattern.eth_match.dst, key->dst, ETH_ALEN);
	}
}

static void tc_flower_store_vlan_match(
				uint32_t index,
				struct flow_dissector_key_vlan *key,
				struct flow_dissector_key_vlan *mask,
				struct qos_mgr_tc_flow_rule_list_item *tc_fl_item)
{
	tc_fl_item->flow_pattern.match_to = QOS_MGR_MATCH_VLAN;

	tc_fl_item->flow_pattern.vlan_match.no_of_tag = index + 1;

	pr_debug("Match Mask vid: %#x pcp: %#x\n", mask->vlan_id, mask->vlan_priority);
	if (mask->vlan_id == 0xfff) {
		pr_debug("<%s><%d> VLAN ID Matched !!!\n",  __func__, __LINE__);
		tc_fl_item->flow_pattern.vlan_match.entry[index].vlan_mask |= QOS_MGR_VLAN_ID;
		tc_fl_item->flow_pattern.vlan_match.entry[index].vid = key->vlan_id;
	}
	if (mask->vlan_priority == 0x7) {
		pr_debug("<%s><%d> VLAN PRIO Matched !!!\n",  __func__, __LINE__);
		tc_fl_item->flow_pattern.vlan_match.entry[index].vlan_mask |= QOS_MGR_VLAN_PRIO;
		tc_fl_item->flow_pattern.vlan_match.entry[index].prio = key->vlan_priority;
	}
/*#ifdef ANYVLAN
	if (tc_fl_item->proto != ETH_P_ANYVLAN) {
		pr_debug("<%s><%d> VLAN TPID Matched !!!\n", __func__, __LINE__);
		tc_fl_item->flow_pattern.vlan_match.entry[index].vlan_mask |= QOS_MGR_VLAN_TPID;
		tc_fl_item->flow_pattern.vlan_match.entry[index].tpid = tc_fl_item->proto;
	}
#endif*/
}

static void tc_flower_store_vlan_proto_match(
				uint32_t index,
				struct flow_dissector_key_basic *key,
				struct flow_dissector_key_basic *mask,
				struct qos_mgr_tc_flow_rule_list_item *tc_fl_item)
{
	tc_fl_item->flow_pattern.match_to = QOS_MGR_MATCH_VLAN;

	pr_debug("Match Mask Proto: %#x\n", mask->n_proto);
	if (mask->n_proto == 0xffff) {
		pr_debug("VLAN Proto Matched !!!\n");
		tc_fl_item->flow_pattern.vlan_match.entry[index].vlan_mask |= QOS_MGR_VLAN_PROTO;
		tc_fl_item->flow_pattern.vlan_match.entry[index].proto = key->n_proto;
	}
}

static void __qos_mgr_tc_fill_vlan_from_pending_flow(
			struct qos_mgr_tc_flow_rule_list_item  *cur_flow,
			struct qos_mgr_tc_flow_rule_list_item *pending_flow)
{
	if (pending_flow->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_ID) {
		cur_flow->flow_pattern.vlan_match.entry[1].vid = pending_flow->flow_pattern.vlan_match.entry[0].vid;
	}
	if (pending_flow->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_PRIO) {
		cur_flow->flow_pattern.vlan_match.entry[1].prio = pending_flow->flow_pattern.vlan_match.entry[0].prio;
	}
	if (pending_flow->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_PROTO) {
		cur_flow->flow_pattern.vlan_match.entry[1].proto = pending_flow->flow_pattern.vlan_match.entry[0].proto;
	}
	if (pending_flow->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_TPID) {
		cur_flow->flow_pattern.vlan_match.entry[1].tpid = pending_flow->flow_pattern.vlan_match.entry[0].tpid;
	}
	cur_flow->flow_pattern.vlan_match.no_of_tag++;

}

int32_t __qos_mgr_tc_find_pending_flow(
				struct qos_mgr_tc_flow_rule_list_item  *fl_list,
				struct qos_mgr_tc_flow_rule_list_item **fl_info)
{
	struct qos_mgr_tc_flow_rule_list_item *p;

	for (p = fl_list; p; p = p->next) {
		if (p->status == 1) {
			*fl_info = p;
			pr_debug("Match Found %p\n", p);
			break;
		}
	}
	return 0;

}

static void qos_mgr_get_device_node_list(void)
{
	int32_t i = 0;
	for (i = 0; i < QOS_MGR_MAX_IFACE; i++) {
		if (tc_qos_mgr_db[i].dev != NULL) {
			pr_debug("Index: %d  Device Name: %s\n", i, tc_qos_mgr_db[i].dev->name);
		}
	}
}

static int32_t qos_mgr_get_free_dev_node_index(void)
{
	int32_t i = 0;
	for (i = 0; i < QOS_MGR_MAX_IFACE; i++) {
		if (tc_qos_mgr_db[i].dev == NULL) {
			pr_err("Dev not found!!! New DB index is %d\n", i);
			break;
		}
	}
	if ((i == (QOS_MGR_MAX_IFACE-1)) || (i > (QOS_MGR_MAX_IFACE-1))) {
		pr_err("DB is FULL\n");
		return PPA_FAILURE;
	}

	return i;
}

static int32_t qos_mgr_get_dev_node_match_index(struct net_device *dev)
{
	int32_t j = 0;
	for (j = 0; j < QOS_MGR_MAX_IFACE; j++) {
		if (tc_qos_mgr_db[j].dev == dev) {
			pr_debug("Matched Device Index is %d\n", j);
			break;
		}
	}
	if ((j == (QOS_MGR_MAX_IFACE-1)) || (j > (QOS_MGR_MAX_IFACE-1)))
		return PPA_FAILURE;

	return j;
}

static int32_t qos_mgr_get_dev_node_index(struct net_device *dev)
{
	int32_t i = 0, index = 0;
	for (i = 0; i < QOS_MGR_MAX_IFACE; i++) {
		if (tc_qos_mgr_db[i].dev == dev) {
			index = i;
			break;
		}
	}
	pr_debug("i= %d\n", i);
	if (i < QOS_MGR_MAX_IFACE)
		goto DEV_FOUND;

	for (i = 0; i < QOS_MGR_MAX_IFACE; i++) {
		if (tc_qos_mgr_db[i].dev == NULL) {
			pr_debug("Dev not found!!! New DB index is %d\n", i);
			break;
		}
	}
	pr_debug("i= %d\n", i);
	index = i;
	if ((i == (QOS_MGR_MAX_IFACE-1)) || (i > (QOS_MGR_MAX_IFACE-1))) {
		pr_err("DB is FULL\n");
		return PPA_FAILURE;
	}
DEV_FOUND:
	return index;

}

#define TC_EGRESS 0
#define TC_INGRESS 1
#define TC_INVALID 2
static int32_t get_clsact_ingress_egress(u32 classid, bool *ingress)
{
	/* This also returns true for ingress qdisc */
	if (TC_H_MAJ(classid) == TC_H_MAJ(TC_H_CLSACT) &&
		TC_H_MIN(classid) != TC_H_MIN(TC_H_MIN_EGRESS))
		*ingress = TC_INGRESS;
	else if (TC_H_MAJ(classid) == TC_H_MAJ(TC_H_CLSACT) &&
		TC_H_MIN(classid) == TC_H_MIN(TC_H_MIN_EGRESS))
		*ingress = TC_EGRESS;
	return 0;
}

/** The qos_mgr_tc_flower_parse function is the main function for parsing the flower rule
	- match the rule based on the flower key and mask
	- store the matched rule in tc_fl_item
	- store the flower actions
*/
static int32_t qos_mgr_tc_flower_parse(struct net_device *dev,
				struct tc_cls_flower_offload *cls_flower,
				bool ingress,
				struct qos_mgr_tc_flow_rule_list_item *tc_fl_item)
{
	struct tc_cls_flower_offload *f = cls_flower;
	const struct tc_action *a;
	int32_t dev_index;
	int32_t indev = (int) *(int *)f->key;
	LIST_HEAD(actions);

	struct net_device *in_dev = dev_get_by_index(&init_net, indev);
	pr_debug("indev: %#x type: %s\n",
			indev, ingress ? "ingress" : "egress");
	/* pr_debug("classid: %#x\n", f->classid); */
	if (in_dev != NULL) {
		pr_debug("Dev: name: %s - ifindex: %d\n",
				in_dev->name, in_dev->ifindex);
	}
	pr_debug("Prio:%d\n", cls_flower->common.prio >> 16);
	pr_debug("Protocol:%02x\n", cls_flower->common.protocol);
	pr_debug("Class Id:%02x\n", cls_flower->common.classid);

	tc_fl_item->dir = ingress;
	tc_fl_item->priority = cls_flower->common.prio >> 16;
	tc_fl_item->proto = cls_flower->common.protocol;
	tc_fl_item->parentid = f->classid;
	tc_fl_item->indev = in_dev;

	if (f->dissector->used_keys &
		~(BIT(FLOW_DISSECTOR_KEY_CONTROL) |
		BIT(FLOW_DISSECTOR_KEY_BASIC) |
		BIT(FLOW_DISSECTOR_KEY_ETH_ADDRS) |
		BIT(FLOW_DISSECTOR_KEY_VLAN) |
		BIT(FLOW_DISSECTOR_KEY_IPV4_ADDRS) |
		BIT(FLOW_DISSECTOR_KEY_IPV6_ADDRS) |
		BIT(FLOW_DISSECTOR_KEY_PORTS))) {
			pr_debug("Unsupported key used: 0x%x\n",
				f->dissector->used_keys);
		}
	pr_debug("Supported key used: 0x%x\n",
				f->dissector->used_keys);

	tc_fl_item->flow_pattern.match_to = QOS_MGR_MATCH_UNKNOWN;
	tc_fl_item->flow_pattern.used_keys = f->dissector->used_keys;

	if (dissector_uses_key(f->dissector, FLOW_DISSECTOR_KEY_CONTROL)) {
		struct flow_dissector_key_control *key =
			skb_flow_dissector_target(f->dissector,
						FLOW_DISSECTOR_KEY_CONTROL,
						f->key);
		pr_debug("Key Addr Type:%d\n", key->addr_type);
	}

	if (dissector_uses_key(f->dissector, FLOW_DISSECTOR_KEY_BASIC)) {
		struct flow_dissector_key_basic *key =
			skb_flow_dissector_target(f->dissector,
						FLOW_DISSECTOR_KEY_BASIC,
						f->key);
		struct flow_dissector_key_basic *mask =
			skb_flow_dissector_target(f->dissector,
						FLOW_DISSECTOR_KEY_BASIC,
						f->mask);
		pr_debug("IP Proto:%d mask:%d\n",
				key->ip_proto, mask->ip_proto);
		pr_debug("Proto:%02x mask:%02x\n",
				key->n_proto, mask->n_proto);
		tc_fl_item->flow_pattern.match_to = QOS_MGR_MATCH_PROTO;
		if (mask->n_proto == 0xffff)
			tc_flower_store_vlan_proto_match(0, key, mask, tc_fl_item);
	}

	if (dissector_uses_key(f->dissector, FLOW_DISSECTOR_KEY_ETH_ADDRS)) {
		struct flow_dissector_key_eth_addrs *key =
			skb_flow_dissector_target(f->dissector,
						FLOW_DISSECTOR_KEY_ETH_ADDRS,
						f->key);
		struct flow_dissector_key_eth_addrs *mask =
			skb_flow_dissector_target(f->dissector,
						FLOW_DISSECTOR_KEY_ETH_ADDRS,
						f->mask);

		pr_debug("MASK src:%x dst:%x\n", mask->src, mask->dst);
		pr_debug("SRC MAC MASK: %02x:%02x:%02x:%02x:%02x:%02x\n",
					mask->src[0], mask->src[1],
					mask->src[2], mask->src[3],
					mask->src[4], mask->src[5]);
		pr_debug("DST MAC MASK: %02x:%02x:%02x:%02x:%02x:%02x\n",
					mask->dst[0], mask->dst[1],
					mask->dst[2], mask->dst[3],
					mask->dst[4], mask->dst[5]);
		pr_debug("SRC MAC: %02x:%02x:%02x:%02x:%02x:%02x\n",
					key->src[0], key->src[1],
					key->src[2], key->src[3],
					key->src[4], key->src[5]);
		pr_debug("DST MAC : %02x:%02x:%02x:%02x:%02x:%02x\n",
					key->dst[0], key->dst[1],
					key->dst[2], key->dst[3],
					key->dst[4], key->dst[5]);
		tc_flower_store_eth_addr_match(key, mask, tc_fl_item);
	}


	/* Classification/Matching arguments parsing */
	if (dissector_uses_key(f->dissector, FLOW_DISSECTOR_KEY_VLAN)) {
		struct flow_dissector_key_vlan *key =
			skb_flow_dissector_target(f->dissector,
					FLOW_DISSECTOR_KEY_VLAN,
					f->key);
		struct flow_dissector_key_vlan *mask =
			skb_flow_dissector_target(f->dissector,
					FLOW_DISSECTOR_KEY_VLAN,
					f->mask);
		pr_debug("Match vid: %#x pcp: %#x\n",
					key->vlan_id, key->vlan_priority);
		tc_flower_store_vlan_match(0, key, mask, tc_fl_item);
	}

	tc_fl_item->status = 0;
	if (g_Rule_status == 1) {
		struct qos_mgr_tc_flow_rule_list_item *fl_info;

		dev_index = qos_mgr_get_dev_node_match_index(dev);
		if (dev_index < QOS_MGR_MAX_IFACE) {
			pr_debug("Dev Index found for ending flow %d !!!\n", dev_index);
			if (ingress == TC_INGRESS)
				__qos_mgr_tc_find_pending_flow(tc_qos_mgr_db[dev_index].flow_item_ig, &fl_info);
			else
				__qos_mgr_tc_find_pending_flow(tc_qos_mgr_db[dev_index].flow_item_eg, &fl_info);

			__qos_mgr_tc_fill_vlan_from_pending_flow(tc_fl_item, fl_info);
			g_Rule_status = 0;
			fl_info->dir = TC_INVALID;
		}
	}

	if (tc_no_actions(f->exts)) {
		pr_err("TC No Action\n");
		return -EINVAL;
	} else {
		pr_debug("No Of Action:%d\n", (f->exts)->nr_actions);
	}
	tcf_exts_to_list(f->exts, &actions);
	list_for_each_entry(a, &actions, list) {
		if (is_tcf_gact_shot(a)) {
			pr_debug("GACT SHOT\n");
			tc_fl_item->flow_action.act_type = TC_ACT_SHOT;
			tc_fl_item->flow_action.action.act = QOS_MGR_ACT_DROP;
			continue;
		} else if (((struct tcf_gact *)a)->tcf_action == TC_ACT_OK) {
			pr_debug("ACTION OK\n");
			continue;
		} else if (((struct tcf_gact *)a)->tcf_action == TC_ACT_UNSPEC) {
			pr_debug("ACTION UNSPEC\n");
			tc_fl_item->flow_action.act_type = TC_ACT_UNSPEC;
			g_Rule_status = 1;
			tc_fl_item->status = 1;
			continue;
		}

		if (is_tcf_mirred_redirect(a)) {
			int ifindex = tcf_mirred_ifindex(a);
			struct net_device *out_dev;

			pr_debug("<%s><%d> Redirect to ifindex:%d\n",
				__func__, __LINE__, ifindex);

			out_dev = __dev_get_by_index(&init_net, ifindex);
			continue;
		}

		if (is_tcf_vlan(a)) {
			if (tcf_vlan_action(a) == TCA_VLAN_ACT_POP) {
				pr_debug("<%s><%d> TCA_VLAN_ACT_POP\n",
					__func__, __LINE__);
				tc_fl_item->flow_action.action.act = QOS_MGR_VLAN_ACT_POP;
				tc_fl_item->flow_action.action.pop_n++;
			} else if (tcf_vlan_action(a) == TCA_VLAN_ACT_PUSH) {
				pr_debug("<%s><%d> TCA_VLAN_ACT_PUSH\n",
					__func__, __LINE__);
				tc_fl_item->flow_action.action.act = QOS_MGR_VLAN_ACT_PUSH;
				tc_fl_item->flow_action.action.vid[tc_fl_item->flow_action.action.push_n] = tcf_vlan_push_vid(a);
				tc_fl_item->flow_action.action.prio[tc_fl_item->flow_action.action.push_n] = ((struct tcf_vlan *)a)->tcfv_push_prio;
				tc_fl_item->flow_action.action.tpid[tc_fl_item->flow_action.action.push_n] = tcf_vlan_push_proto(a);
				tc_fl_item->flow_action.action.push_n++;
			} else { /* action is TCA_VLAN_ACT_MODIFY */
				pr_debug("<%s><%d> TCA_VLAN_ACT_MODIFY\n",
					__func__, __LINE__);
				tc_fl_item->flow_action.action.act = QOS_MGR_VLAN_ACT_MODIFY;
				tc_fl_item->flow_action.action.vid[0] = tcf_vlan_push_vid(a);
				tc_fl_item->flow_action.action.prio[0] = ((struct tcf_vlan *)a)->tcfv_push_prio;
				tc_fl_item->flow_action.action.tpid[0] = tcf_vlan_push_proto(a);
			}
			continue;
		}
	}
	return 0;
}

/** The qos_mgr_setup_tc function is the main function for TC HW offload
	- this is implemented with support for the tc hooks using the kernel ndo_setup_tc ops
	- extracting direction(ingress/egress), priority, protocol
	- offload the extended VLAN rule of the flower command
	- collect all VLAN information for this CTP or brige port based on 
	- specified direction Ingress or Egress
	- for every netdevice two linked lists are maintained per direction
	- Based on the netdevice and direction, rules are added to either ingress/egress linked list 
*/
int32_t qos_mgr_setup_tc(struct net_device *dev,
			u32 handle,
			__be16 protocol,
			struct tc_to_netdev *tc)
{
	int32_t ret = PPA_SUCCESS;

	pr_debug("TC Setup for device %s === Handle: %x Protocol: %x\n",
				dev->name, handle, protocol);
	if (tc->type == TC_SETUP_CLSFLOWER) {
		switch (tc->cls_flower->command) {
			bool ingress;
			int32_t dev_index;
		case TC_CLSFLOWER_REPLACE:
			{
				struct qos_mgr_tc_flow_rule_list_item *tc_fl_item;
				pr_debug("TC_CLSFLOWER_REPLACE\n");
				get_clsact_ingress_egress(
					tc->cls_flower->common.classid,
					&ingress);
				dev_index = qos_mgr_get_dev_node_match_index(dev);
				if (dev_index == PPA_FAILURE) {
					pr_err("No match found !!!\n");
					dev_index = qos_mgr_get_free_dev_node_index();
					if (dev_index == PPA_FAILURE)
						return -1;
				}
				pr_debug("New device index in DB : %d dev pointer %p name %s\n", dev_index, dev, dev->name);
				tc_qos_mgr_db[dev_index].dev = dev;
				tc_fl_item = tc_qos_mgr_alloc_list_item();
				if (!tc_fl_item)
					pr_err("Obj error\n");

				qos_mgr_tc_flower_parse(dev, tc->cls_flower, ingress, tc_fl_item);
				if (ingress) {
					__qos_mgr_tc_add_list_item(tc_fl_item, &(tc_qos_mgr_db[dev_index].flow_item_ig));
					qos_mgr_set_ext_vlan_to_dp(dev, tc_qos_mgr_db[dev_index].flow_item_ig);
				} else {
					__qos_mgr_tc_add_list_item(tc_fl_item, &(tc_qos_mgr_db[dev_index].flow_item_eg));
					qos_mgr_set_ext_vlan_to_dp(dev, tc_qos_mgr_db[dev_index].flow_item_eg);
				}

				break;
			}
		case TC_CLSFLOWER_DESTROY:
			{
				uint32_t prio;
				pr_debug("TC_CLSFLOWER_DESTROY\n");
				dev_index = qos_mgr_get_dev_node_match_index(dev);
				if (dev_index == PPA_FAILURE) {
					return -1;
				} else {
					prio = tc->cls_flower->common.prio >> 16;
					get_clsact_ingress_egress(tc->cls_flower->common.classid, &ingress);
					if (ingress) {
						tc_qos_mgr_remove_list_item(prio, &(tc_qos_mgr_db[dev_index].flow_item_ig));
						qos_mgr_set_ext_vlan_to_dp(dev, tc_qos_mgr_db[dev_index].flow_item_ig);
					} else {
						tc_qos_mgr_remove_list_item(prio, &(tc_qos_mgr_db[dev_index].flow_item_eg));
						qos_mgr_set_ext_vlan_to_dp(dev, tc_qos_mgr_db[dev_index].flow_item_eg);
					}
				}
				break;
			}
		case TC_CLSFLOWER_STATS:
			{
				pr_debug("TC_CLSFLOWER_STATS\n");
				dev_index = qos_mgr_get_dev_node_match_index(dev);
				if (dev_index == PPA_FAILURE) {
					return -1;
				} else {
					get_clsact_ingress_egress(tc->cls_flower->common.classid, &ingress);
					if (ingress)
						qos_mgr_tc_flow_lookup(tc_qos_mgr_db[dev_index].flow_item_ig);
					else
						qos_mgr_tc_flow_lookup(tc_qos_mgr_db[dev_index].flow_item_eg);
				}
				break;
			}
		default:
			return -EINVAL;
		}

	}
	return ret;
}

void qos_mgr_tc_init(void)
{
	int32_t i = 0, j = 0;
	for (i = 0; i < QOS_MGR_MAX_IFACE; i++) {
		tc_qos_mgr_db[i].dev = NULL;
		tc_qos_mgr_db[i].flow_item_ig = NULL;
		tc_qos_mgr_db[i].flow_item_eg = NULL;
		for (j = 0; j < QOS_MGR_MAX_IFACE; j++) {
			memset(&tc_qos_mgr_db[i].mq_info, 0, sizeof(struct qos_mgr_tc_queue_info));
		}
	}
	pr_debug("<%s> Init Done.\n", __func__);
}

static int32_t qos_mgr_find_no_of_vlan_rules(
				struct qos_mgr_tc_flow_rule_list_item  *fl_list,
				int32_t *vlan0,
				int32_t *vlan1,
				int32_t *vlan2)
{
	int32_t i = 0, j = 0, k = 0;
	struct qos_mgr_tc_flow_rule_list_item *p;
	for (p = fl_list; p; p = p->next) {
		if (p->flow_pattern.match_to == QOS_MGR_MATCH_PROTO)
			i++;
		else if ((p->flow_pattern.match_to == QOS_MGR_MATCH_VLAN) && (p->dir != TC_INVALID)) {
			if (p->flow_pattern.vlan_match.no_of_tag == 1)
				j++;
			else if (p->flow_pattern.vlan_match.no_of_tag == 2)
				k++;
		}
	}
	*vlan0 = i;
	*vlan1 = j;
	*vlan2 = k;
	return 0;
}

static void qos_mgr_show_dp_info(struct dp_tc_vlan *dp_vlan)
{
	int32_t i = 0;

	pr_debug("netdevice: %s\n", dp_vlan->dev->name);
	pr_debug("Direction: %d\n", dp_vlan->dir);
	pr_debug("==== No of No vlan tag:%d Single vlan:%d Double vlan:%d ====\n",
			dp_vlan->n_vlan0,
			dp_vlan->n_vlan1,
			dp_vlan->n_vlan2);
	for (i = 0; i < dp_vlan->n_vlan0; i++) {
		pr_debug("Pattern: Proto:%d\n", dp_vlan->vlan0_list->outer.proto);
	}
	for (i = 0; i < dp_vlan->n_vlan1; i++) {
		pr_debug("Single VLAN\n");
		pr_debug("Pattern: Proto:%d\n", dp_vlan->vlan1_list->outer.proto);
		pr_debug("Pattern: Vid:%d\n", dp_vlan->vlan1_list->outer.vid);
		pr_debug("Pattern: Prio:%d\n", dp_vlan->vlan1_list->outer.prio);
	}
	for (i = 0; i < dp_vlan->n_vlan2; i++) {
		pr_debug("Double VLAN\n");
		pr_debug("Pattern Outer: Proto:%d\n", dp_vlan->vlan2_list->outer.proto);
		pr_debug("Pattern: Vid:%d\n", dp_vlan->vlan2_list->outer.vid);
		pr_debug("Pattern: Prio:%d\n", dp_vlan->vlan2_list->outer.prio);

		pr_debug("Pattern: Inner Proto:%d\n", dp_vlan->vlan2_list->inner.proto);
		pr_debug("Pattern: Vid:%d\n", dp_vlan->vlan2_list->inner.vid);
		pr_debug("Pattern: Prio:%d\n", dp_vlan->vlan2_list->inner.prio);
	}
}

/** The qos_mgr_set_ext_vlan_to_dp function is used to get all the vlan rules of any 
	direction and then call dp_vlan_set() API 
*/
static int32_t qos_mgr_set_ext_vlan_to_dp(
				struct net_device *dev,
				struct qos_mgr_tc_flow_rule_list_item  *fl_list)
{
	int32_t ret = PPA_SUCCESS;
	struct qos_mgr_tc_flow_rule_list_item *p;

	struct dp_tc_vlan dp_vlan = {0};
	struct dp_vlan0 *rule0_head = NULL;
	struct dp_vlan1 *rule1_head = NULL;
	struct dp_vlan2 *rule2_head = NULL;
	int32_t i = 0;
	int32_t vlan0_index = 0, vlan1_index = 0, vlan2_index = 0;

	dp_vlan.dev = dev;
	dp_vlan.dir = fl_list->dir;

	qos_mgr_find_no_of_vlan_rules(fl_list, &dp_vlan.n_vlan0, &dp_vlan.n_vlan1, &dp_vlan.n_vlan2);

	if (dp_vlan.n_vlan0) {
		rule0_head = (struct dp_vlan0 *)ppa_malloc(sizeof(struct dp_vlan0) * dp_vlan.n_vlan0);
		if (!rule0_head) {
			pr_err("Can't allocate memory\n");
			ret = PPA_FAILURE;
			goto END;
		}
	}
	if (dp_vlan.n_vlan1) {
		rule1_head = (struct dp_vlan1 *)ppa_malloc(sizeof(struct dp_vlan1) * dp_vlan.n_vlan1);
		if (!rule1_head) {
			pr_err("Can't allocate memory\n");
			ret = PPA_FAILURE;
			goto END;
		}
	}
	if (dp_vlan.n_vlan2) {
		rule2_head = (struct dp_vlan2 *)ppa_malloc(sizeof(struct dp_vlan2) * dp_vlan.n_vlan2);
		if (!rule2_head) {
			pr_err("Can't allocate memory\n");
			ret = PPA_FAILURE;
			goto END;
		}
	}

	for (p = fl_list; p; p = p->next) {
		if ((p->flow_pattern.match_to == QOS_MGR_MATCH_VLAN) && (p->dir != TC_INVALID))
			pr_debug("<%s> No Of VLAN Tag %d\n", __func__,
					p->flow_pattern.vlan_match.no_of_tag);

		if (p->flow_pattern.match_to == QOS_MGR_MATCH_PROTO) {
			struct dp_vlan0 *rule0 = NULL;
			rule0 = rule0_head + vlan0_index;

			rule0->outer.proto = p->proto;
			rule0->outer.prio = DP_VLAN_PATTERN_NOT_CARE;
			rule0->outer.tpid = DP_VLAN_PATTERN_NOT_CARE;

			if (p->flow_action.action.act == QOS_MGR_ACT_DROP) {
				rule0->def = DP_VLAN_ACT_DROP;
			}
			rule0->act.act = DP_VLAN_ACT_PUSH;
			for (i = 0; i < p->flow_action.action.push_n; i++) {
				rule0->act.tpid[i] = p->flow_action.action.tpid[i];
				rule0->act.vid[i] = p->flow_action.action.vid[i];
				rule0->act.prio[i] = p->flow_action.action.prio[i];
			}
			rule0->act.push_n = p->flow_action.action.push_n;
			vlan0_index++;
		} else if ((p->flow_pattern.match_to == QOS_MGR_MATCH_VLAN) && (p->dir != TC_INVALID)) {
			if (p->flow_pattern.vlan_match.no_of_tag == 1) {
				struct dp_vlan1 *rule1 = NULL;
				rule1 = rule1_head + vlan1_index;

				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_ID) ?
						(rule1->outer.vid = p->flow_pattern.vlan_match.entry[0].vid) : (rule1->outer.vid = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_PRIO) ?
						(rule1->outer.prio = p->flow_pattern.vlan_match.entry[0].prio) : (rule1->outer.prio = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_PROTO) ?
						(rule1->outer.proto = p->flow_pattern.vlan_match.entry[0].proto) : (rule1->outer.proto = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_TPID) ?
						(rule1->outer.tpid = p->flow_pattern.vlan_match.entry[0].tpid) : (rule1->outer.tpid = DP_VLAN_PATTERN_NOT_CARE);

				if (p->flow_action.action.act == QOS_MGR_ACT_DROP) {
					rule1->def = DP_VLAN_ACT_DROP;
				} else if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_POP) {
					rule1->act.act = DP_VLAN_ACT_POP;
					rule1->act.push_n = p->flow_action.action.pop_n;
				} else if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_PUSH) {
					for (i = 0; i < p->flow_action.action.push_n; i++) {
						rule1->act.tpid[i] = p->flow_action.action.tpid[i];
						rule1->act.vid[i] = p->flow_action.action.vid[i];
						rule1->act.prio[i] = p->flow_action.action.prio[i];
					}
					rule1->act.act = DP_VLAN_ACT_PUSH;
					rule1->act.push_n = p->flow_action.action.push_n;
				}
				vlan1_index++;
			} else if (p->flow_pattern.vlan_match.no_of_tag == 2) {
				struct dp_vlan2 *rule2 = NULL;
				rule2 = rule2_head + vlan2_index;

				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_ID) ?
						(rule2->outer.vid = p->flow_pattern.vlan_match.entry[0].vid) : (rule2->outer.vid = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_PRIO) ?
						(rule2->outer.prio = p->flow_pattern.vlan_match.entry[0].prio) : (rule2->outer.prio = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_PROTO) ?
						(rule2->outer.proto = p->flow_pattern.vlan_match.entry[0].proto) : (rule2->outer.proto = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[0].vlan_mask & QOS_MGR_VLAN_TPID) ?
						(rule2->outer.tpid = p->flow_pattern.vlan_match.entry[0].tpid) : (rule2->outer.tpid = DP_VLAN_PATTERN_NOT_CARE);

				(p->flow_pattern.vlan_match.entry[1].vlan_mask & QOS_MGR_VLAN_ID) ?
						(rule2->inner.vid = p->flow_pattern.vlan_match.entry[1].vid) : (rule2->inner.vid = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[1].vlan_mask & QOS_MGR_VLAN_PRIO) ?
						(rule2->inner.prio = p->flow_pattern.vlan_match.entry[1].prio) : (rule2->inner.prio = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[1].vlan_mask & QOS_MGR_VLAN_PROTO) ?
						(rule2->inner.proto = p->flow_pattern.vlan_match.entry[1].proto) : (rule2->inner.proto = DP_VLAN_PATTERN_NOT_CARE);
				(p->flow_pattern.vlan_match.entry[1].vlan_mask & QOS_MGR_VLAN_TPID) ?
						(rule2->inner.tpid = p->flow_pattern.vlan_match.entry[1].tpid) : (rule2->inner.tpid = DP_VLAN_PATTERN_NOT_CARE);

				if (p->flow_action.action.act == QOS_MGR_ACT_DROP) {
					rule2->def = DP_VLAN_ACT_DROP;
				} else if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_POP) {
					rule2->act.act = DP_VLAN_ACT_POP;
					rule2->act.push_n = p->flow_action.action.pop_n;
				} else if (p->flow_action.action.act == QOS_MGR_VLAN_ACT_PUSH) {
					for (i = 0; i < p->flow_action.action.push_n; i++) {
						rule2->act.tpid[i] = p->flow_action.action.tpid[i];
						rule2->act.vid[i] = p->flow_action.action.vid[i];
						rule2->act.prio[i] = p->flow_action.action.prio[i];
					}
					rule2->act.act = DP_VLAN_ACT_PUSH;
					rule2->act.push_n = p->flow_action.action.push_n;
				}
				vlan2_index++;
			}
		}
		dp_vlan.vlan0_list = rule0_head;
		dp_vlan.vlan1_list = rule1_head;
		dp_vlan.vlan2_list = rule2_head;

		qos_mgr_show_dp_info(&dp_vlan);
		if(dp_vlan_set(&dp_vlan, 0) == DP_FAILURE)
			ret = PPA_FAILURE;
		
		ret = PPA_SUCCESS;
	}
END:
	if (rule0_head)
		ppa_free(rule0_head);
	if (rule1_head)
		ppa_free(rule1_head);
	if (rule2_head)
		ppa_free(rule2_head);

	return ret;
}
